@@ -15,7 +15,8 @@
 #  limitations under the License.
 #
 # ISOFIT: Imaging Spectrometer Optimal FITting
-# Author: David R Thompson, david.r.thompson@jpl.nasa.gov
+# Authors: David R Thompson, david.r.thompson@jpl.nasa.gov
+#          Nimrod Carmon, nimrod.carmon@jpl.nasa.gov
 #
 
 from sys import platform
@@ -41,6 +42,7 @@ import subprocess
 ### Variables ###
 
 eps = 1e-5  # used for finite difference derivative calculations
+tropopause_altitude_km = 17.
 
 ### Classes ###
 
@@ -90,11 +92,24 @@ class ModtranRT(TabularRT):
             self.aer_extc = np.array(aer_extc)
             self.aer_asym = np.array(aer_asym)
 
+        # Determine whether we are using the three run or single run strategy
+        self.multipart_transmittance = engine_config.multipart_transmittance
+
+        # Idenfity the physical quantities we will calculate
         self.modtran_lut_names = ['rhoatm', 'transm', 'sphalb', 'transup']
+
+        # Special emissive terms
         if self.treat_as_emissive:
             self.modtran_lut_names = ['thermal_upwelling',
                                       'thermal_downwelling'] + self.modtran_lut_names
 
+        # If excercising the multipart transmittance option we will run with
+        # three reflectance values
+        if self.multipart_transmittance:
+            self.test_rfls = [0, 0.1, 0.5]
+            self.modtran_lut_names = self.modtran_lut_names + \
+                  ['t_down_dir', 't_down_dif', 't_up_dir', 't_up_dif']
+
         self.last_point_looked_up = np.zeros(self.n_point)
         self.last_point_lookup_values = np.zeros(self.n_point)
 
@@ -151,6 +166,14 @@ class ModtranRT(TabularRT):
              * transm  - diffuse and direct irradiance along the
                           sun-ground-sensor path
              * transup - transmission along the ground-sensor path only
+             
+           If the "multipart transmittance" option is active, we will use
+           a combination of three MODTRAN runs to estimate the following 
+           additional quantities:
+             * t_down_dir - direct downwelling transmittance
+             * t_down_dif - diffuse downwelling transmittance
+             * t_up_dir   - direct upwelling transmittance
+             * t_up_dif   - diffuse upwelling transmittance
 
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
             Be careful with these! They are to be used only by the
@@ -169,20 +192,40 @@ class ModtranRT(TabularRT):
         with open(infile) as f:
             sols, transms, sphalbs, wls, rhoatms, transups = \
                 [], [], [], [], [], []
+            t_down_dirs, t_down_difs, t_up_dirs, t_up_difs = [],[],[],[]
+            grnd_rflts_1, drct_rflts_1, grnd_rflts_2, drct_rflts_2 = \
+                [], [], [], []
+            transm_dirs, transm_difs, widths = [],[],[]
+            lp_0, lp_1, lp_2 = [],[],[]
             thermal_upwellings, thermal_downwellings = [], []
             lines = f.readlines()
+            nheader = 5
+
+            # Mark header and data segments
+            nwl = len(self.wl)
+            case = -np.ones(nheader*3+nwl*3)
+            case[nheader:(nheader+nwl)] = 0
+            case[(nheader*2+nwl):(nheader*2+nwl*2)] = 1
+            case[(nheader*3+nwl*2):(nheader*3+nwl*3)] = 2
+
             for i, line in enumerate(lines):
-                if i < 5:
+
+                # exclude headers
+                if case[i] < 0:
                     continue
+                    
+                # parse data out of each line in the MODTRAN output
                 toks = line.strip().split(' ')
                 toks = re.findall(r"[\S]+", line.strip())
                 wl, wid = float(toks[0]), float(toks[8])  # nm
                 solar_irr = float(toks[18]) * 1e6 * \
                     np.pi / wid / coszen  # uW/nm/sr/cm2
-                rdnatm = float(toks[4]) * 1e6  # uW/nm/sr/cm2
-                rhoatm = rdnatm * np.pi / (solar_irr * coszen)
-                sphalb = float(toks[23])
-                transm = float(toks[22]) + float(toks[21])
+                rdnatm  = float(toks[4]) * 1e6  # uW/nm/sr/cm2
+                rhoatm  = rdnatm * np.pi / (solar_irr * coszen)
+                sphalb  = float(toks[23])
+                A_coeff = float(toks[21]) 
+                B_coeff = float(toks[22])
+                transm  = A_coeff + B_coeff 
                 transup = float(toks[24])
 
                 # Be careful with these! See note in function comments above
@@ -193,21 +236,99 @@ class ModtranRT(TabularRT):
 
                 # Be careful with these! See note in function comments above
                 # grnd_rflt already includes ground-to-sensor transmission
-                grnd_rflt = float(toks[16])
-                thermal_downwelling = grnd_rflt / wid * 1e6  # uW/nm/sr/cm2
+                grnd_rflt = float(toks[16]) * 1e6  # ground reflected radiance (direct+diffuse+multiple scattering)
+                drct_rflt = float(toks[17]) * 1e6  # same as 16 but only on the sun->surface->sensor path (only direct)
+                path_rdn  = float(toks[14]) * 1e6 + float(toks[14]) * 1e6  # The sum of the (1) single scattering and (2) multiple scattering 
+                thermal_downwelling = grnd_rflt / wid # uW/nm/sr/cm2
 
-                sols.append(solar_irr)
-                transms.append(transm)
-                sphalbs.append(sphalb)
-                rhoatms.append(rhoatm)
-                transups.append(transup)
+                if case[i] == 0:
 
-                thermal_upwellings.append(thermal_upwelling)
-                thermal_downwellings.append(thermal_downwelling)
+                     sols.append(solar_irr)      # solar irradiance 
+                     transms.append(transm)      # total transmittance 
+                     sphalbs.append(sphalb)      # spherical albedo 
+                     rhoatms.append(rhoatm)      # atmospheric reflectance 
+                     transups.append(transup)    # upwelling direct transmittance 
+                     transm_dirs.append(A_coeff) # total direct transmittance 
+                     transm_difs.append(B_coeff) # total diffuse transmittance 
+                     widths.append(wid)          # channel width in nm 
+                     lp_0.append(path_rdn)       # path radiance of zero surface reflectance 
+                     thermal_upwellings.append(thermal_upwelling)
+                     thermal_downwellings.append(thermal_downwelling)
+                     wls.append(wl) #wavelengths in nm 
+
+                elif case[i] == 1:
+
+                     grnd_rflts_1.append(grnd_rflt) #total ground reflected radiance 
+                     drct_rflts_1.append(drct_rflt) #direct path ground reflected radiance 
+                     lp_1.append(path_rdn) #path radiance (sum of single and multiple scattering)
+
+                elif case[i] == 2:
+
+                     grnd_rflts_2.append(grnd_rflt) #total ground reflected radiance 
+                     drct_rflts_2.append(drct_rflt) #direct path ground reflected radiance 
+                     lp_2.append(path_rdn) #path radiance (sum of single and multiple scattering)
+
+        if self.multipart_transmittance: 
+            ''' 
+                This implementation is following Gaunter et al. (2009) (DOI:10.1080/01431160802438555),
+                and modified by Nimrod Carmon. It is called the "2-albedo" method, referring to running 
+                modtran with 2 different surface albedos. The 3-albedo method is similar to this one with 
+                the single difference where the "path_radiance_no_surface" variable is taken from a
+                zero-surface-reflectance modtran run instead of being calculated from 2 modtran outputs.
+                There are a few argument as to why this approach is beneficial:
+                (1) for each grid point on the lookup table you sample modtran 2 or 3 times, i.e. you get 
+                2 or 3 "data points" for the atmospheric parameter of interest. This in theory allows us 
+                to use a lower band model resolution modtran run, which is much faster, while keeping 
+                high accuracy. Currently we have the 5 cm-1 band model resolution configured.
+                The second advantage is the possibility to use the decoupled transmittance products to exapnd 
+                the forward model and account for more physics e.g. shadows \ sky view \ adjacency \ terrain etc.
+                  
+            '''
+            t_up_dirs = np.array(transups) 
+            direct_ground_reflected_1   = np.array(drct_rflts_1) 
+            total_ground_reflected_1    = np.array(grnd_rflts_1) 
+            direct_ground_reflected_2   = np.array(drct_rflts_2) 
+            total_ground_reflected_2    = np.array(grnd_rflts_2) 
+            path_radiance_1       = np.array(lp_1) 
+            path_radiance_2       = np.array(lp_2) 
+            TOA_Irad   = np.array(sols) * coszen / np.pi
+            rfl_1      = self.test_rfls[1]
+            rfl_2      = self.test_rfls[2]
+            mus        = coszen
+            
+
+            direct_flux_1 = direct_ground_reflected_1 * np.pi / rfl_1 / t_up_dirs 
+            global_flux_1 = total_ground_reflected_1 * np.pi / rfl_1 / t_up_dirs 
+            diffuse_flux_1 = global_flux_1 - direct_flux_1 # diffuse flux
+
+            global_flux_2 = total_ground_reflected_2 * np.pi / rfl_2 / t_up_dirs
+
+            path_radiance_no_surface = (rfl_2 * path_radiance_1 * global_flux_2 - \
+                                        rfl_1 * path_radiance_2 * global_flux_1) / \
+                            (rfl_2 * global_flux_2 - rfl_1 * global_flux_1)
+
+            # Diffuse upwelling transmittance
+            t_up_difs =  np.pi * (path_radiance_1 - path_radiance_no_surface) / \
+                                 (rfl_1 * global_flux_1) 
+
+            # Spherical Albedo
+            sphalbs = (global_flux_1 - global_flux_2) / \
+                      (rfl_1 * global_flux_1 - rfl_2 * global_flux_2)
+            direct_flux_radiance = direct_flux_1/mus
+
+            global_flux_no_surface = global_flux_1*(1.-rfl_1 * sphalbs) 
+            diffuse_flux_no_surface = global_flux_no_surface - direct_flux_radiance * coszen
+            
+            t_down_dirs = (direct_flux_radiance * coszen / wid / np.pi) / TOA_Irad
+            t_down_difs = (diffuse_flux_no_surface / wid / np.pi) / TOA_Irad
+            
+            # total transmittance
+            transms = (t_down_dirs + t_down_difs) * (t_up_dirs + t_up_difs)
+            
+        params = [np.array(i) for i in [wls, sols, rhoatms, transms, sphalbs, transups,
+                                        t_down_dirs, t_down_difs, t_up_dirs, t_up_difs, 
+                                        thermal_upwellings, thermal_downwellings]]
 
-                wls.append(wl)
-        params = [np.array(i) for i in [wls, sols, rhoatms, transms, sphalbs,
-                                        transups, thermal_upwellings, thermal_downwellings]]
         return tuple(params)
 
     def ext550_to_vis(self, ext550):
@@ -258,6 +379,60 @@ class ModtranRT(TabularRT):
             # Geometry parameters we want to populate even if unassigned
             elif key in ['H1ALT', 'IDAY', 'TRUEAZ','OBSZEN', 'GMTIME' ]:
                 param[0]['MODTRANINPUT']['GEOMETRY'][key] = val
+            
+            elif key == 'AIRT_DELTA_K':
+                # If there is no profile already provided ...
+                if param[0]['MODTRANINPUT']['ATMOSPHERE']['MODEL'] != "ATM_USER_ALT_PROFILE":
+
+                    # MODTRAN cannot accept a ground altitude above 6 km, so keep all layers after that
+                    gndalt = param[0]['MODTRANINPUT']['SURFACE']['GNDALT']
+
+                    # E.g.: [1.5, 2, 3, 4, 5]
+                    low_altitudes = [gndalt] + list(np.arange(6 - np.ceil(gndalt)) + np.ceil(gndalt))
+
+                    # MODTRAN cannot accept a ground altitude above 6 km, so keep all layers after that
+                    hi_altitudes = [6.0,7.0,8.0,9.0,10.0,11.0,12.0,13.0,14.0,15.0,16.0,17.0,18.0,19.0,
+                                    20.0,21.0,22.0,23.0,24.0,25.0,30.0,35.0,40.0,45.0,50.0,55.0,60.0,70.0,80.0,100.0] 
+
+                    altitudes = low_altitudes + hi_altitudes # Append lists, don't add altitudes!
+
+                    prof_unt_tdelta_kelvin = np.where(np.array(altitudes) <= tropopause_altitude_km, val, 0)
+
+                    altitude_dict = {
+                        'TYPE': 'PROF_ALTITUDE', 'UNITS': 'UNT_KILOMETERS', 'PROFILE': altitudes}
+                    delta_kelvin_dict = {
+                        'TYPE': 'PROF_TEMPERATURE', 'UNITS': 'UNT_TDELTA_KELVIN', 'PROFILE': prof_unt_tdelta_kelvin.tolist() }
+
+                    param[0]['MODTRANINPUT']['ATMOSPHERE']['MODEL'] = "ATM_USER_ALT_PROFILE"
+                    param[0]['MODTRANINPUT']['ATMOSPHERE']['NPROF'] = 2
+                    param[0]['MODTRANINPUT']['ATMOSPHERE']['NLAYERS'] = len(altitudes)
+                    param[0]['MODTRANINPUT']['ATMOSPHERE']['PROFILES'] = [altitude_dict, delta_kelvin_dict]
+
+                
+                else: # A profile is already provided, assume that it includes PROF_ALTITUDE
+                    nprof = param[0]['MODTRANINPUT']['ATMOSPHERE']['NPROF']
+                    profile_types = []
+                    for i in range(nprof):
+                        profile_types.append(param[0]['MODTRANINPUT']['ATMOSPHERE']['PROFILES'][i]['TYPE'])
+
+                    ind_prof_altitude = profile_types.index('PROF_ALTITUDE')
+                    prof_altitude = np.array(param[0]['MODTRANINPUT']['ATMOSPHERE']['PROFILES'][ind_prof_altitude]['PROFILE'])
+
+                    if 'PROF_TEMPERATURE' in profile_types:
+                        # If a temperature profile already exists, then we must add the temperature delta to that
+                        # as MODTRAN apparently does not allow have both an offset and a specified temperature
+                        ind_prof_temperature = profile_types.index('PROF_TEMPERATURE')
+                        prof_temperature = np.array(param[0]['MODTRANINPUT']['ATMOSPHERE']['PROFILES'][ind_prof_temperature]['PROFILE'])
+                        prof_temperature = np.where(prof_altitude <= tropopause_altitude_km, prof_temperature + val, prof_temperature)
+                        param[0]['MODTRANINPUT']['ATMOSPHERE']['PROFILES'][ind_prof_temperature]['PROFILE'] = prof_temperature.tolist()
+                    
+                    else:
+                        # If a temperature profile does not exist, then use UNT_TDELTA_KELVIN
+                        prof_unt_tdelta_kelvin = np.where(prof_altitude <= tropopause_altitude_km, val, 0.)
+                        prof_unt_tdelta_kelvin_dict = {
+                            'TYPE': 'PROF_TEMPERATURE', 'UNITS': 'UNT_TDELTA_KELVIN', 'PROFILE': prof_unt_tdelta_kelvin.tolist() }
+                        param[0]['MODTRANINPUT']['ATMOSPHERE']['PROFILES'].append(prof_unt_tdelta_kelvin_dict)
+                        param[0]['MODTRANINPUT']['ATMOSPHERE']['NPROF'] = nprof + 1
 
             # Surface parameters we want to populate even if unassigned
             elif key in ['GNDALT']:
@@ -299,6 +474,19 @@ class ModtranRT(TabularRT):
             lvl0['EXTC'] = [float(v) / total_extc550 for v in total_extc]
             lvl0['ABSC'] = [float(v) / total_extc550 for v in total_absc]
 
+        if self.multipart_transmittance:
+            # Here we copy the original config and just change the surface reflectance 
+            param[0]['MODTRANINPUT']['CASE'] = 0
+            param[0]['MODTRANINPUT']['SURFACE']['SURREF']= self.test_rfls[0]
+            param1 = deepcopy(param[0])
+            param1['MODTRANINPUT']['CASE'] = 1
+            param1['MODTRANINPUT']['SURFACE']['SURREF']= self.test_rfls[1]
+            param.append(param1)
+            param2 = deepcopy(param[0]) 
+            param2['MODTRANINPUT']['CASE'] = 2
+            param2['MODTRANINPUT']['SURFACE']['SURREF']= self.test_rfls[2]
+            param.append(param2)
+
         return json.dumps({"MODTRAN": param}), param
 
     def build_lut(self, rebuild=False):
@@ -468,7 +656,8 @@ class ModtranRT(TabularRT):
         # the modtran_tir functions as they require the modtran reflectivity
         # be set to 1 in order to use them in the RTM in radiative_transfer.py.
         # Don't add these to the VSWIR functions!
-        names = ['wl', 'sol', 'rhoatm', 'transm', 'sphalb', 'transup']
+        names = ['wl', 'sol', 'rhoatm', 'transm', 'sphalb', 'transup',
+                't_down_dir','t_down_dif','t_up_dir','t_up_dif']
 
         # Don't include the thermal terms in VSWIR runs to avoid incorrect usage
         if self.treat_as_emissive:
