@@ -11,7 +11,7 @@ from datetime import datetime
 from spectral.io import envi
 import logging
 import json
-import gdal
+from osgeo import gdal
 import numpy as np
 from sklearn import mixture
 import subprocess
@@ -27,7 +27,7 @@ SEGMENTATION_SIZE = 400
 
 UNCORRELATED_RADIOMETRIC_UNCERTAINTY = 0.01
 
-INVERSION_WINDOWS = [[400.0, 1300.0], [1450, 1780.0], [2050.0, 2450.0]]
+INVERSION_WINDOWS = [[380.0, 1340.0], [1450, 1800.0], [1970.0, 2500.0]]
 
 
 def main():
@@ -93,7 +93,7 @@ def main():
     parser.add_argument('input_loc', type=str)
     parser.add_argument('input_obs', type=str)
     parser.add_argument('working_directory', type=str)
-    parser.add_argument('sensor', type=str, choices=['ang', 'avcl', 'neon', 'prism'])
+    parser.add_argument('sensor', type=str)
     parser.add_argument('--copy_input_files', type=int, choices=[0,1], default=0)
     parser.add_argument('--modtran_path', type=str)
     parser.add_argument('--wavelength_path', type=str)
@@ -116,6 +116,11 @@ def main():
 
     args = parser.parse_args()
 
+    if args.sensor not in ['ang', 'avcl', 'neon', 'prism', 'emit']:
+        if args.sensor[:3] != 'NA-':
+            raise ValueError('argument sensor: invalid choice: "NA-test" (choose from '
+                             '"ang", "avcl", "neon", "prism", "emit", "NA-*")')
+
     if args.copy_input_files == 1:
         args.copy_input_files = True
     else:
@@ -127,6 +132,11 @@ def main():
         logging.basicConfig(format='%(message)s', level=args.logging_level, filename=args.log_file)
 
     lut_params = LUTConfig(args.lut_config_file)
+    if args.emulator_base is not None:
+        lut_params.aot_550_range = lut_params.aerosol_2_range
+        lut_params.aot_550_spacing = lut_params.aerosol_2_spacing
+        lut_params.aot_550_spacing_min = lut_params.aerosol_2_spacing_min
+        lut_params.aerosol_2_spacing = 0
 
     paths = Pathnames(args)
     paths.make_directories()
@@ -137,17 +147,19 @@ def main():
     if args.sensor == 'ang':
         # parse flightline ID (AVIRIS-NG assumptions)
         dt = datetime.strptime(paths.fid[3:], '%Y%m%dt%H%M%S')
-        dayofyear = dt.timetuple().tm_yday
     elif args.sensor == 'avcl':
         # parse flightline ID (AVIRIS-CL assumptions)
         dt = datetime.strptime('20{}t000000'.format(paths.fid[1:7]), '%Y%m%dt%H%M%S')
-        dayofyear = dt.timetuple().tm_yday
     elif args.sensor == 'neon':
         dt = datetime.strptime(paths.fid, 'NIS01_%Y%m%d_%H%M%S')
-        dayofyear = dt.timetuple().tm_yday
     elif args.sensor == 'prism':
         dt = datetime.strptime(paths.fid[3:], '%Y%m%dt%H%M%S')
-        dayofyear = dt.timetuple().tm_yday
+    elif args.sensor == 'emit':
+        dt = datetime.strptime(paths.fid[:19], 'emit%Y%m%dt%H%M%S')
+    elif args.sensor[:3] == 'NA-':
+        dt = datetime.strptime(args.sensor[3:], '%Y%m%d')
+
+    dayofyear = dt.timetuple().tm_yday
 
     h_m_s, day_increment, mean_path_km, mean_to_sensor_azimuth, mean_to_sensor_zenith, valid, \
     to_sensor_azimuth_lut_grid, to_sensor_zenith_lut_grid = get_metadata_from_obs(paths.obs_working_path, lut_params)
@@ -183,13 +195,24 @@ def main():
                               fwhm[:, np.newaxis]], axis=1)
     np.savetxt(paths.wavelength_path, wl_data, delimiter=' ')
 
-    mean_latitude, mean_longitude, mean_elevation_km, elevation_lut_grid = get_metadata_from_loc(paths.loc_working_path, lut_params)
+    mean_latitude, mean_longitude, mean_elevation_km, elevation_lut_grid = \
+        get_metadata_from_loc(paths.loc_working_path, lut_params)
+    if args.emulator_base is not None:
+        if elevation_lut_grid is not None and np.any(elevation_lut_grid < 0):
+            to_rem = elevation_lut_grid[elevation_lut_grid < 0].copy()
+            elevation_lut_grid[ elevation_lut_grid< 0] = 0
+            elevation_lut_grid = np.unique(elevation_lut_grid)
+            logging.info("Scene contains target lut grid elements < 0 km, and uses 6s (via sRTMnet).  6s does not "
+                         f"support targets below sea level in km units.  Setting grid points {to_rem} to 0.")
 
     # Need a 180 - here, as this is already in MODTRAN convention
     mean_altitude_km = mean_elevation_km + np.cos(np.deg2rad(180 - mean_to_sensor_zenith)) * mean_path_km
 
-    logging.info('Path (km): %f, 180 - To-sensor Zenith (deg): %f, To-sensor Azimuth (deg) : %f, Altitude: %6.2f km' %
-                 (mean_path_km, mean_to_sensor_zenith, mean_to_sensor_azimuth, mean_altitude_km))
+    logging.info('Observation means:')
+    logging.info(f'Path (km): {mean_path_km}')
+    logging.info(f'To-sensor Zenith (deg): {mean_to_sensor_zenith}')
+    logging.info(f'To-sensor Azimuth (deg): {mean_to_sensor_azimuth}')
+    logging.info(f'Altitude (km): {mean_altitude_km}')
 
     # We will use the model discrepancy with covariance OR uncorrelated 
     # Calibration error, but not both.
@@ -203,7 +226,8 @@ def main():
         if not exists(paths.lbl_working_path) or not exists(paths.radiance_working_path):
             logging.info('Segmenting...')
             segment(spectra=(paths.radiance_working_path, paths.lbl_working_path),
-                    nodata_value=-9999, npca=5, segsize=SEGMENTATION_SIZE, nchunk=CHUNKSIZE, n_cores=args.n_cores)
+                    nodata_value=-9999, npca=5, segsize=SEGMENTATION_SIZE, nchunk=CHUNKSIZE,
+                    n_cores=args.n_cores)
 
         # Extract input data per segment
         for inp, outp in [(paths.radiance_working_path, paths.rdn_subs_path),
@@ -212,8 +236,7 @@ def main():
             if not exists(outp):
                 logging.info('Extracting ' + outp)
                 extractions(inputfile=inp, labels=paths.lbl_working_path,
-                            output=outp, chunksize=CHUNKSIZE, flag=-9999)
-
+                            output=outp, chunksize=CHUNKSIZE, flag=-9999, n_cores=args.n_cores)
 
     if args.presolve == 1:
 
@@ -233,7 +256,7 @@ def main():
         if not exists(paths.h2o_subs_path + '.hdr') or not exists(paths.h2o_subs_path):
             # Write the presolve connfiguration file
             h2o_grid = np.linspace(0.01, max_water - 0.01, 10).round(2)
-            logging.info('Pre-solve H2O grid: {}'.format(h2o_grid))
+            logging.info(f'Pre-solve H2O grid: {h2o_grid}')
             logging.info('Writing H2O pre-solve configuration file.')
             build_presolve_config(paths, h2o_grid, args.n_cores, args.empirical_line == 1, args.surface_category,
                 args.emulator_base, uncorrelated_radiometric_uncertainty)
@@ -258,23 +281,17 @@ def main():
         p05 = np.percentile(h2o_est[h2o_est > lut_params.h2o_min], 5)
         p95 = np.percentile(h2o_est[h2o_est > lut_params.h2o_min], 95)
         margin = (p95-p05) * 0.25
-        h2o_lo = max(lut_params.h2o_min, p05 - margin)
-        h2o_hi = min(max_water, max(lut_params.h2o_min, p95 + margin))
-        h2o_lut_grid = np.linspace(h2o_lo, h2o_hi,
-                lut_params.num_h2o_lut_elements)
 
-        if (np.abs(h2o_lut_grid[-1] - h2o_lut_grid[0]) < 0.03):
-            new_h2o_lut_grid = np.linspace(h2o_lut_grid[0] - 0.1* np.ceil(lut_params.num_h2o_lut_elements/2.), 
-                                           h2o_lut_grid[0] + 0.1*np.ceil(lut_params.num_h2o_lut_elements/2.), 
-                                           lut_params.num_h2o_lut_elements)
-            logging.warning('Warning: h2o lut grid from presolve detected as {}-{}, which is very narrow.  Expanding to {}-{}.  Advised to check presolve solutions thoroughly.'.format(h2o_lut_grid[0],h2o_lut_grid[-1], new_h2o_lut_grid[0], new_h2o_lut_grid[-1]))
-            h2o_lut_grid = new_h2o_lut_grid
-    else:
-        h2o_lut_grid = np.linspace(lut_params.default_h2o_lut_range[0], lut_params.default_h2o_lut_range[1],
-                                   lut_params.num_h2o_lut_elements)
+        lut_params.h2o_range[0] = max(lut_params.h2o_min, p05 - margin)
+        lut_params.h2o_range[1] = min(max_water, max(lut_params.h2o_min, p95 + margin))
 
+    h2o_lut_grid = lut_params.get_grid(lut_params.h2o_range[0], lut_params.h2o_range[1], lut_params.h2o_spacing, lut_params.h2o_spacing_min)
 
-    logging.info('Full (non-aerosol) LUTs:\nElevation: {}\nTo-sensor azimuth: {}\nTo-sensor zenith: {}\nh2o-vis: {}:'.format(elevation_lut_grid, to_sensor_azimuth_lut_grid, to_sensor_zenith_lut_grid, h2o_lut_grid))
+    logging.info('Full (non-aerosol) LUTs:')
+    logging.info(f'Elevation: {elevation_lut_grid}')
+    logging.info(f'To-sensor azimuth: {to_sensor_azimuth_lut_grid}')
+    logging.info(f'To-sensor zenith: {to_sensor_zenith_lut_grid}')
+    logging.info(f'H2O Vapor: {h2o_lut_grid}')
 
     logging.info(paths.state_subs_path)
     if not exists(paths.state_subs_path) or \
@@ -296,6 +313,7 @@ def main():
         logging.info('Running ISOFIT with full LUT')
         retrieval_full = isofit.Isofit(paths.modtran_config_path, level='INFO', logfile=args.log_file)
         retrieval_full.run()
+        del retrieval_full
 
         # clean up unneeded storage
         for to_rm in ['*r_k', '*t_k', '*tp7', '*wrn', '*psc', '*plt', '*7sc', '*acd']:
@@ -341,6 +359,10 @@ class Pathnames():
             logging.info('Flightline ID: %s' % self.fid)
         elif args.sensor == 'neon':
             self.fid = split(args.input_radiance)[-1][:21]
+        elif args.sensor == 'emit':
+            self.fid = split(args.input_radiance)[-1][:19]
+        elif args.sensor[3:] == 'NA-':
+            self.fid = os.path.splitext(os.path.basename(args.input_radiance))[0]
 
         # Names from inputs
         self.aerosol_climatology = args.aerosol_climatology_path
@@ -437,7 +459,10 @@ class Pathnames():
         self.irradiance_file = abspath(join(self.isofit_path,'examples','20151026_SantaMonica','data','prism_optimized_irr.dat'))
 
         self.aerosol_tpl_path = join(self.isofit_path, 'data', 'aerosol_template.json')
-        self.rdn_factors_path = args.rdn_factors_path
+        self.rdn_factors_path = None
+        if args.rdn_factors_path is not None:
+            self.rdn_factors_path = abspath(args.rdn_factors_path)
+
 
         self.ray_temp_dir = args.ray_temp_dir
 
@@ -497,29 +522,55 @@ class LUTConfig:
             with open(lut_config_file, 'r') as f:
                 lut_config = json.load(f)
 
-        self.num_elev_lut_elements = 1
-        self.num_h2o_lut_elements = 3
-        self.num_to_sensor_azimuth_lut_elements = 1
-        self.num_to_sensor_zenith_lut_elements = 1
+        # For each element, set the look up table spacing (lut_spacing) as the
+        # anticipated spacing value, or 0 to use a single point (not LUT).
+        # Set the 'lut_spacing_min' as the minimum distance allowed - if separation
+        # does not meet this threshold based on the available data, on a single
+        # point will be used.
 
-        # Setting any of these to '1' will also remove that aerosol from the statevector
-        self.num_aerosol_1_lut_elements = 1
-        self.num_aerosol_2_lut_elements = 3
-        self.num_aerosol_3_lut_elements = 3
+        # Units of kilometers
+        self.elevation_spacing = 0.5
+        self.elevation_spacing_min = 0.2
 
-        self.aerosol_1_lut_range = [0.001, 0.5]
-        self.aerosol_2_lut_range = [0.001, 0.5]
-        self.aerosol_3_lut_range = [0.001, 0.5]
-
-        self.aot_550_lut_range = [0.001, 0.5]
-        self.num_aot_550_elements = 1
+        # Units of g / m2
+        self.h2o_spacing = 0.25
+        self.h2o_spacing_min = 0.03
 
+        # Special parameter to specify the minimum allowable water vapor value in g / m2
         self.h2o_min = 0.2
 
-        self.zenith_min_spacing = 2
+        # Set defaults, will override based on settings
+        # Units of g / m2
+        self.h2o_range = [0.05, 5]
 
-        self.default_h2o_lut_range = [0.05, 5]
+        # Units of degrees
+        self.to_sensor_azimuth_spacing = 0
+        self.to_sensor_azimuth_spacing_min = 0
 
+        # Units of degrees
+        self.to_sensor_zenith_spacing = 10
+        self.to_sensor_zenith_spacing_min = 2
+
+        # Units of AOD
+        self.aerosol_0_spacing = 0
+        self.aerosol_0_spacing_min = 0
+
+        # Units of AOD
+        self.aerosol_1_spacing = 0
+        self.aerosol_1_spacing_min = 0
+
+        # Units of AOD
+        self.aerosol_2_spacing = 0.25
+        self.aerosol_2_spacing_min = 0
+
+        # Units of AOD
+        self.aerosol_0_range = [0.001, 0.5]
+        self.aerosol_1_range = [0.001, 0.5]
+        self.aerosol_2_range = [0.001, 0.5]
+        self.aot_550_range = [0.001, 0.5]
+
+        self.aot_550_spacing = 0
+        self.aot_550_spacing_min = 0
 
         # overwrite anything that comes in from the config file
         if lut_config_file is not None:
@@ -527,6 +578,114 @@ class LUTConfig:
                 if key in self.__dict__:
                     setattr(self, key, lut_config[key])
 
+    def get_grid(self, minval: float, maxval: float, spacing: float, min_spacing: float):
+        if spacing == 0:
+            logging.debug('Grid spacing set at 0, using no grid.')
+            return None
+        num_gridpoints = int(np.ceil((maxval-minval)/spacing)) + 1
+
+        grid = np.linspace(minval, maxval, num_gridpoints)
+
+        if min_spacing > 0.0001:
+            grid = np.round(grid, 4)
+        if len(grid) == 1:
+            logging.debug(f'Grid spacing is 0, which is less than {min_spacing}.  No grid used')
+            return None
+        elif np.abs(grid[1] - grid[0]) < min_spacing:
+            logging.debug(f'Grid spacing is {grid[1]-grid[0]}, which is less than {min_spacing}.  No grid used')
+            return None
+        else:
+            return grid
+
+    def get_angular_grid(self, angle_data_input: np.array, spacing: float, min_spacing: float, units : str = 'd'):
+        """ Find either angular data 'center points' (num_points = 1), or a lut set that spans
+        angle variation in a systematic fashion.
+
+        Args:
+            angle_data_input: set of angle data to use to find center points
+            spacing: the desired angular spacing between points, or mean if -1
+            min_spacing: the minimum angular spacing between points allowed (if less, no grid)
+            units: specifies if data are in degrees (default) or radians
+
+        :Returns:
+            angular data center point or lut set spanning space
+
+        """
+        if spacing == 0:
+            logging.debug('Grid spacing set at 0, using no grid.')
+            return None
+
+        # Convert everything to radians so we don't have to track throughout
+        if units == 'r':
+            angle_data = np.rad2deg(angle_data_input)
+        else:
+            angle_data = angle_data_input.copy()
+
+        spatial_data = np.hstack([np.cos(np.deg2rad(angle_data)).reshape(-1, 1),
+                                  np.sin(np.deg2rad(angle_data)).reshape(-1, 1)])
+
+        # find which quadrants have data
+        quadrants = np.zeros((2, 2))
+        if np.any(np.logical_and(spatial_data[:, 0] > 0, spatial_data[:, 1] > 0)):
+            quadrants[1, 0] = 1
+        if np.any(np.logical_and(spatial_data[:, 0] > 0, spatial_data[:, 1] < 0)):
+            quadrants[1, 1] += 1
+        if np.any(np.logical_and(spatial_data[:, 0] < 0, spatial_data[:, 1] > 0)):
+            quadrants[0, 0] += 1
+        if np.any(np.logical_and(spatial_data[:, 0] < 0, spatial_data[:, 1] < 0)):
+            quadrants[0, 1] += 1
+
+        # Handle the case where angles are < 180 degrees apart
+        if np.sum(quadrants) < 3 and spacing != -1:
+            if np.sum(quadrants[1, :]) == 2:
+                # If angles cross the 0-degree line:
+                angle_spread = self.get_grid(np.min(angle_data + 180), np.max(angle_data + 180), spacing, min_spacing)
+                if angle_spread is None:
+                    return None
+                else:
+                    return angle_spread - 180
+            else:
+                # Otherwise, just space things out:
+                return self.get_grid(np.min(angle_data), np.max(angle_data), spacing, min_spacing)
+        else:
+            if spacing >= 180:
+                logging.warning(f'Requested angle spacing is {spacing}, but obs angle divergence is > 180.  '
+                                'Tighter  spacing recommended')
+
+            # If we're greater than 180 degree spread, there's no universal answer. Try GMM.
+            if spacing == -1:
+                num_points = 1
+            else:
+                # This very well might overly space the grid, but we don't / can't know in general
+                num_points = int(np.ceil(360 / spacing))
+
+            # We initialize the GMM with a static seed for repeatability across runs
+            gmm = mixture.GaussianMixture(n_components=num_points, covariance_type='full',
+                                          random_state=1)
+            if spatial_data.shape[0]  == 1:
+                spatial_data = np.vstack([spatial_data, spatial_data])
+
+            gmm.fit(spatial_data)
+            central_angles = np.degrees(np.arctan2(gmm.means_[:, 1], gmm.means_[:, 0]))
+            if num_points == 1:
+                return central_angles[0]
+
+            ca_quadrants = np.zeros((2, 2))
+            if np.any(np.logical_and(gmm.means_[:, 0] > 0, gmm.means_[:, 1] > 0)):
+                ca_quadrants[1, 0] = 1
+            elif np.any(np.logical_and(gmm.means_[:, 0] > 0, gmm.means_[:, 1] < 0)):
+                ca_quadrants[1, 1] += 1
+            elif np.any(np.logical_and(gmm.means_[:, 0] < 0, gmm.means_[:, 1] > 0)):
+                ca_quadrants[0, 0] += 1
+            elif np.any(np.logical_and(gmm.means_[:, 0] < 0, gmm.means_[:, 1] < 0)):
+                ca_quadrants[0, 1] += 1
+
+            if np.sum(ca_quadrants) < np.sum(quadrants):
+                logging.warning(f'GMM angles {central_angles} span {np.sum(ca_quadrants)} quadrants, '
+                                f'while data spans {np.sum(ca_quadrants)} quadrants')
+
+            return central_angles
+
 
 def load_climatology(config_path: str, latitude: float, longitude: float, acquisition_datetime: datetime,
                      isofit_path: str, lut_params: LUTConfig) -> (np.array, np.array, np.array):
@@ -551,10 +710,12 @@ def load_climatology(config_path: str, latitude: float, longitude: float, acquis
     aerosol_model_path = join(isofit_path, 'data', 'aerosol_model.txt')
     aerosol_state_vector = {}
     aerosol_lut_grid = {}
-    aerosol_lut_ranges = [lut_params.aerosol_1_lut_range, lut_params.aerosol_2_lut_range, lut_params.aerosol_2_lut_range]
-    num_aerosol_lut_elements = [lut_params.num_aerosol_1_lut_elements, lut_params.num_aerosol_2_lut_elements, lut_params.num_aerosol_3_lut_elements]
+    aerosol_lut_ranges = [lut_params.aerosol_0_range, lut_params.aerosol_1_range, lut_params.aerosol_2_range]
+    aerosol_lut_spacing = [lut_params.aerosol_0_spacing, lut_params.aerosol_1_spacing, lut_params.aerosol_2_spacing]
+    aerosol_lut_spacing_mins = [lut_params.aerosol_0_spacing_min, lut_params.aerosol_1_spacing_min, lut_params.aerosol_2_spacing_min]
     for _a, alr in enumerate(aerosol_lut_ranges):
-        if num_aerosol_lut_elements[_a] != 1:
+        aerosol_lut = lut_params.get_grid(alr[0], alr[1], aerosol_lut_spacing[_a], aerosol_lut_spacing_mins[_a])
+        if aerosol_lut is not None:
             aerosol_state_vector['AERFRAC_{}'.format(_a)] = {
                 "bounds": [float(alr[0]), float(alr[1])],
                 "scale": 1,
@@ -562,12 +723,12 @@ def load_climatology(config_path: str, latitude: float, longitude: float, acquis
                 "prior_sigma": 10.0,
                 "prior_mean": float((alr[1] - alr[0]) / 10. + alr[0])}
 
-            aerosol_lut = np.linspace(alr[0], alr[1], num_aerosol_lut_elements[_a])
-            aerosol_lut_grid['AERFRAC_{}'.format(_a)] = [round(float(q),4) for q in aerosol_lut]
+            aerosol_lut_grid['AERFRAC_{}'.format(_a)] = aerosol_lut.tolist()
 
-    if lut_params.num_aot_550_elements != 1:
-        aerosol_lut = np.linspace(lut_params.aot_550_lut_range[0], lut_params.aot_550_lut_range[1], lut_params.num_aot_550_elements)
-        aerosol_lut_grid['AOT550'] = [round(float(q),4) for q in aerosol_lut]
+    aot_550_lut = lut_params.get_grid(lut_params.aot_550_range[0], lut_params.aot_550_range[1],
+                                      lut_params.aot_550_spacing, lut_params.aot_550_spacing_min)
+    if aot_550_lut is not None:
+        aerosol_lut_grid['AOT550'] = aot_550_lut.tolist()
         alr = [aerosol_lut_grid['AOT550'][0], aerosol_lut_grid['AOT550'][-1]]
         aerosol_state_vector['AOT550'] = {
                         "bounds": [float(alr[0]), float(alr[1])],
@@ -576,7 +737,6 @@ def load_climatology(config_path: str, latitude: float, longitude: float, acquis
                         "prior_sigma": 10.0,
                         "prior_mean": float((alr[1] - alr[0]) / 10. + alr[0])}
 
-
     logging.info('Loading Climatology')
     # If a configuration path has been provided, use it to get relevant info
     if config_path is not None:
@@ -661,82 +821,6 @@ def calc_modtran_max_water(paths: Pathnames) -> float:
     return max_water
 
 
-def find_angular_seeds(angle_data_input: np.array, num_points: int, units: str = 'd') -> np.array:
-    """ Find either angular data 'center points' (num_points = 1), or a lut set that spans
-    angle variation in a systematic fashion.
-
-    Args:
-        angle_data_input: set of angle data to use to find center points
-        num_points: the number of points to find - if 1, this will return a single point (the 'centerpoint'), if > 1,
-                    this will return a numpy array spanning the specified number of poitns
-        units: specifies if data are in degrees (default) or radians
-
-    :Returns:
-        central_angle - angular data center point or lut set spanning space
-
-    """
-
-    # Convert everything to radians so we don't have to track throughout
-    if units == 'r':
-        angle_data = np.rad2deg(angle_data_input)
-    else:
-        angle_data = angle_data_input.copy()
-
-    spatial_data = np.hstack([np.cos(np.deg2rad(angle_data)).reshape(-1, 1),
-                              np.sin(np.deg2rad(angle_data)).reshape(-1, 1)])
-
-    # find which quadrants have data
-    quadrants = np.zeros((2,2))
-    if np.any(np.logical_and(spatial_data[:,0] > 0, spatial_data[:,1] > 0 )):
-        quadrants[1,0] = 1
-    if np.any(np.logical_and(spatial_data[:,0] > 0, spatial_data[:,1] < 0 )):
-        quadrants[1,1] += 1
-    if np.any(np.logical_and(spatial_data[:,0] < 0, spatial_data[:,1] > 0 )):
-        quadrants[0,0] += 1
-    if np.any(np.logical_and(spatial_data[:,0] < 0, spatial_data[:,1] < 0 )):
-        quadrants[0,1] += 1
-
-    # Handle the case where angles are < 180 degrees apart
-    if np.sum(quadrants) < 3 and num_points != 1:
-        if (np.sum(quadrants[1,:]) == 2):
-            # If angles cross the 0-degree line:
-            angle_spread = np.linspace(np.min(angle_data+180), np.max(angle_data+180), num_points) - 180
-            return angle_spread
-        else:
-            # Otherwise, just space things out:
-            return np.linspace(np.min(angle_data), np.max(angle_data), num_points)
-    else:
-        # If we're greater than 180 degree spread, there's no universal answer. Try GMM.
-
-        if num_points == 2:
-            logging.warning('2 angle interpolation selected when angle divergence > 180. '
-                            'At least 3 points are recommended')
-
-        # We initialize the GMM with a static seed for repeatability across runs
-        gmm = mixture.GaussianMixture(n_components=num_points, covariance_type='full',
-                random_state=1)
-        gmm.fit(spatial_data)
-        central_angles = np.degrees(np.arctan2(gmm.means_[:,1], gmm.means_[:,0]))
-        if (num_points == 1):
-            return central_angles[0]
-
-        ca_quadrants = np.zeros((2, 2))
-        if np.any(np.logical_and(gmm.means_[:, 0] > 0, gmm.means_[:, 1] > 0)):
-            ca_quadrants[1, 0] = 1
-        elif np.any(np.logical_and(gmm.means_[:, 0] > 0, gmm.means_[:, 1] < 0)):
-            ca_quadrants[1, 1] += 1
-        elif np.any(np.logical_and(gmm.means_[:, 0] < 0, gmm.means_[:, 1] > 0)):
-            ca_quadrants[0, 0] += 1
-        elif np.any(np.logical_and(gmm.means_[:, 0] < 0, gmm.means_[:, 1] < 0)):
-            ca_quadrants[0, 1] += 1
-
-        if np.sum(ca_quadrants) < np.sum(quadrants):
-            logging.warning('GMM angles {} span {} quadrants, while data spans {} quadrants'.format(central_angles,
-                            np.sum(ca_quadrants), np.sum(quadrants)))
-
-        return central_angles
-
-
 def get_metadata_from_obs(obs_file: str, lut_params: LUTConfig, trim_lines: int = 5,
                           max_flight_duration_h: int = 8, nodata_value: float = -9999) -> \
                           (List, bool, float, float, float, np.array, List, List):
@@ -795,21 +879,17 @@ def get_metadata_from_obs(obs_file: str, lut_params: LUTConfig, trim_lines: int
     mean_path_km = np.mean(path_km[valid])
     del path_km
 
-    mean_to_sensor_azimuth = find_angular_seeds(to_sensor_azimuth[valid], 1) % 360
-    mean_to_sensor_zenith = 180 - find_angular_seeds(to_sensor_zenith[valid], 1)
+    mean_to_sensor_azimuth = lut_params.get_angular_grid(to_sensor_azimuth[valid], -1, 0) % 360
+    mean_to_sensor_zenith = 180 - lut_params.get_angular_grid(to_sensor_zenith[valid], -1, 0)
 
     #geom_margin = EPS * 2.0
-    if lut_params.num_to_sensor_zenith_lut_elements == 1:
-        to_sensor_zenith_lut_grid = None
-    else:
-        to_sensor_zenith_lut_grid = np.sort(180 - find_angular_seeds(to_sensor_zenith[valid], lut_params.num_to_sensor_zenith_lut_elements))
-        if (to_sensor_zenith_lut_grid[1] - to_sensor_zenith_lut_grid[0] < lut_params.zenith_min_spacing):
-            to_sensor_zenith_lut_grid = None
+    to_sensor_zenith_lut_grid = lut_params.get_angular_grid(to_sensor_zenith[valid], lut_params.to_sensor_zenith_spacing, lut_params.to_sensor_zenith_spacing_min)
+    if to_sensor_zenith_lut_grid is not None:
+        to_sensor_zenith_lut_grid = np.sort(180 - to_sensor_zenith_lut_grid)
 
-    if lut_params.num_to_sensor_azimuth_lut_elements == 1:
-        to_sensor_azimuth_lut_grid = None
-    else:
-        to_sensor_azimuth_lut_grid = np.sort(np.array([x % 360 for x in find_angular_seeds(to_sensor_azimuth[valid], lut_params.num_to_sensor_azimuth_lut_elements)]))
+    to_sensor_azimuth_lut_grid = lut_params.get_angular_grid(to_sensor_azimuth[valid], lut_params.to_sensor_azimuth_spacing, lut_params.to_sensor_azimuth_spacing_min)
+    if to_sensor_azimuth_lut_grid is not None:
+        to_sensor_azimuth_lut_grid = np.sort(np.array([x % 360 for x in to_sensor_azimuth_lut_grid]))
 
     del to_sensor_azimuth
     del to_sensor_zenith
@@ -877,25 +957,20 @@ def get_metadata_from_loc(loc_file: str, lut_params: LUTConfig, trim_lines: int
         valid[-trim_lines:, :] = False
 
     # Grab zensor position and orientation information
-    mean_latitude = find_angular_seeds(loc_data[1,valid].flatten(),1)
-    mean_longitude = find_angular_seeds(-1 * loc_data[0,valid].flatten(),1)
+    mean_latitude = lut_params.get_angular_grid(loc_data[1,valid].flatten(), -1, 0)
+    mean_longitude = lut_params.get_angular_grid(-1 * loc_data[0,valid].flatten(), -1, 0)
 
     mean_elevation_km = np.mean(loc_data[2,valid]) / 1000.0
 
     # make elevation grid
-    if lut_params.num_elev_lut_elements == 1:
-        elevation_lut_grid = None
-    else:
-        min_elev = np.min(loc_data[2, valid])/1000.
-        max_elev = np.max(loc_data[2, valid])/1000.
-        elevation_lut_grid = np.linspace(max(min_elev, EPS),
-                                         max_elev,
-                                         lut_params.num_elev_lut_elements)
+    min_elev = np.min(loc_data[2, valid]) / 1000.
+    max_elev = np.max(loc_data[2, valid]) / 1000.
+    elevation_lut_grid = lut_params.get_grid(min_elev, max_elev, lut_params.elevation_spacing,
+                                             lut_params.elevation_spacing_min)
 
     return mean_latitude, mean_longitude, mean_elevation_km, elevation_lut_grid
 
 
-
 def build_presolve_config(paths: Pathnames, h2o_lut_grid: np.array, n_cores: int=-1,
         use_emp_line:bool = False, surface_category="multicomponent_surface",
         emulator_base: str = None, uncorrelated_radiometric_uncertainty: float = 0.0):
@@ -1058,21 +1133,22 @@ def build_main_config(paths: Pathnames, lut_params: LUTConfig, h2o_lut_grid: np.
                     #statevector_names - populated below
                 }
             },
-            "statevector": {
-                "H2OSTR": {
-                    "bounds": [h2o_lut_grid[0], h2o_lut_grid[-1]],
-                    "scale": 1,
-                    "init": (h2o_lut_grid[1] + h2o_lut_grid[-1]) / 2.0,
-                    "prior_sigma": 100.0,
-                    "prior_mean": (h2o_lut_grid[1] + h2o_lut_grid[-1]) / 2.0,
-                }
-            },
+            "statevector": {},
             "lut_grid": {},
             "unknowns": {
                 "H2O_ABSCO": 0.0
             }
     }
 
+    if h2o_lut_grid is not None:
+        radiative_transfer_config['statevector']['H2OSTR'] = {
+            "bounds": [h2o_lut_grid[0], h2o_lut_grid[-1]],
+            "scale": 1,
+            "init": (h2o_lut_grid[1] + h2o_lut_grid[-1]) / 2.0,
+            "prior_sigma": 100.0,
+            "prior_mean": (h2o_lut_grid[1] + h2o_lut_grid[-1]) / 2.0,
+        }
+
     if emulator_base is not None:
         radiative_transfer_config['radiative_transfer_engines']['vswir']['emulator_file'] = abspath(emulator_base)
         radiative_transfer_config['radiative_transfer_engines']['vswir']['emulator_aux_file'] = abspath(emulator_base + '_aux.npz')
@@ -1086,13 +1162,13 @@ def build_main_config(paths: Pathnames, lut_params: LUTConfig, h2o_lut_grid: np.
 
 
     if h2o_lut_grid is not None:
-        radiative_transfer_config['lut_grid']['H2OSTR'] = [round(max(0.0, float(q)),4) for q in h2o_lut_grid]
+        radiative_transfer_config['lut_grid']['H2OSTR'] = h2o_lut_grid.tolist()
     if elevation_lut_grid is not None:
-        radiative_transfer_config['lut_grid']['GNDALT'] = [round(max(0.0, float(q)),4) for q in elevation_lut_grid]
+        radiative_transfer_config['lut_grid']['GNDALT'] = elevation_lut_grid.tolist()
     if to_sensor_azimuth_lut_grid is not None:
-        radiative_transfer_config['lut_grid']['TRUEAZ'] = [round(float(q),4) for q in to_sensor_azimuth_lut_grid]
+        radiative_transfer_config['lut_grid']['TRUEAZ'] = to_sensor_azimuth_lut_grid.tolist()
     if to_sensor_zenith_lut_grid is not None:
-        radiative_transfer_config['lut_grid']['OBSZEN'] = [round(float(q),4) for q in to_sensor_zenith_lut_grid] # modtran convension
+        radiative_transfer_config['lut_grid']['OBSZEN'] = to_sensor_zenith_lut_grid.tolist() # modtran convension
 
     # add aerosol elements from climatology
     aerosol_state_vector, aerosol_lut_grid, aerosol_model_path = \
@@ -1143,13 +1219,18 @@ def build_main_config(paths: Pathnames, lut_params: LUTConfig, h2o_lut_grid: np.
 
     if multiple_restarts:
         eps = 1e-2
-        h2o_delta = float(h2o_lut_grid[-1]) - float(h2o_lut_grid[0])
-        grid = {'H2OSTR':[round(h2o_lut_grid[0]+h2o_delta*0.02,4), round(h2o_lut_grid[-1]-h2o_delta*0.02,4)]}
+        grid = {}
+        if h2o_lut_grid is not None:
+            h2o_delta = float(h2o_lut_grid[-1]) - float(h2o_lut_grid[0])
+            grid['H2OSTR'] = [round(h2o_lut_grid[0]+h2o_delta*0.02,4), 
+                              round(h2o_lut_grid[-1]-h2o_delta*0.02,4)]
+
         # We will initialize using different AODs for the first aerosol in the LUT
         if len(aerosol_lut_grid)>0:
             key = list(aerosol_lut_grid.keys())[0]
             aer_delta = aerosol_lut_grid[key][-1] - aerosol_lut_grid[key][0]
-            grid[key] = [round(aerosol_lut_grid[key][0]+aer_delta*0.02), round(aerosol_lut_grid[key][-1]-aer_delta*0.02,4)]
+            grid[key] = [round(aerosol_lut_grid[key][0]+aer_delta*0.02,4), 
+                         round(aerosol_lut_grid[key][-1]-aer_delta*0.02,4)]
         isofit_config_modtran['implementation']['inversion']['integration_grid'] = grid
         isofit_config_modtran['implementation']['inversion']['inversion_grid_as_preseed'] = True
 
